# -*- coding: utf-8 -*-

from b3j0f.conf import Configurable, category, Parameter

from link.middleware.core import register_middleware
from link.kvstore.driver import Driver

from link.crdt.core import CRDT

from link.riak.features.crdt import convert_crdt_from_riak
from link.riak.features.crdt import convert_crdt_to_riak
from link.riak.features.fulltext import RiakSearch2
from link.riak import CONF_BASE_PATH
from link.feature import addfeatures

from riak.security import SecurityCreds
from riak import RiakClient, RiakObject
from riak.datatypes import Datatype

from six import raise_from


@Configurable(
    paths='{0}/driver.conf'.format(CONF_BASE_PATH),
    conf=category(
        'RIAK',
        Parameter(name='default_bucket', value='default'),
        Parameter(name='protocol', value='http'),
        Parameter(name='pkey'),
        Parameter(name='cert'),
        Parameter(name='cacert'),
        Parameter(name='crl'),
        Parameter(name='ciphers'),
        Parameter(name='sslver', ptype=int)
    )
)
@addfeatures([RiakSearch2])
@register_middleware
class RiakDriver(Driver):

    __protocols__ = ['riak']

    def __init__(
        self,
        default_bucket=None,
        protocol=None,
        pkey=None,
        cert=None,
        cacert=None,
        crl=None,
        ciphers=None,
        sslver=5,
        *args, **kwargs
    ):
        super(RiakDriver, self).__init__(*args, **kwargs)

        self.default_bucket = default_bucket
        self.protocol = protocol
        self.pkey = pkey
        self.cert = cert
        self.cacert = cacert
        self.crl = crl
        self.ciphers = ciphers
        self.sslver = sslver

    def _connect(self):
        nodes = []
        security = None

        for host, port in self.hosts:
            node = {
                'host': host
            }

            if self.protocol == 'pbc':
                node['pb_port'] = port

            else:
                node['http_port'] = port

            nodes.append(node)

        kwargs = {}

        if self.user is not None:
            kwargs['username'] = self.user

            if self.pwd:
                kwargs['password'] = self.pwd

        if self.pkey is not None:
            kwargs['pkey_file'] = self.pkey

        if self.cert is not None:
            kwargs['cert_file'] = self.cert

        if self.cacert is not None:
            kwargs['cacert_file'] = self.cacert

        if self.crl is not None:
            kwargs['crl_file'] = self.crl

        if self.ciphers is not None:
            kwargs['ciphers'] = self.ciphers

        if kwargs:
            kwargs['ssl_version'] = self.sslver

            security = SecurityCreds(**kwargs)

        else:
            security = None

        return RiakClient(nodes=nodes, credentials=security)

    def _disconnect(self, conn):
        conn.close()

    def _isconnected(self, conn):
        return conn is not None and conn.is_alive()

    def _get_bucket(self, conn):
        if self.path is None or len(self.path) == 0:
            bucket = conn.bucket(self.default_bucket)

        elif len(self.path) == 1:
            bucket = conn.bucket(str(self.path[0]))

        else:
            bucket_type = conn.bucket_type(str(self.path[0]))
            bucket = bucket_type.bucket('_'.join([
                str(p)
                for p in self.path[1:]
            ]))

        return bucket

    def _new_object(self, conn, key, val):
        if not isinstance(val, CRDT):
            obj = self._get_bucket(conn).new(key, val)
            return obj

        else:
            crdt = val

        riak_crdt = convert_crdt_to_riak(crdt)
        riak_crdt.bucket = self._get_bucket(conn)
        riak_crdt.key = key
        riak_crdt.update()

    def _get(self, conn, key):
        key = str(key)

        obj = self._get_bucket(conn).get(key)

        if isinstance(obj, RiakObject):
            if not obj.exists:
                raise KeyError('No such key: {0}'.format(key))

            return obj.data

        elif isinstance(obj, Datatype):
            return convert_crdt_from_riak(obj)

        else:
            raise KeyError('No such key: {0}'.format(key))

    def _multiget(self, conn, keys):
        keys = list(map(str, keys))
        bucket = self._get_bucket(conn)

        results = bucket.multiget(keys)

        datas = []

        for result in results:
            if isinstance(result, RiakObject):
                datas.append(result.data)

            elif isinstance(result, Datatype):
                doc = convert_crdt_from_riak(result)
                datas.append(doc)

            else:
                _, _, key, err = result

                raise_from(
                    KeyError('No such key: {0}'.format(key)),
                    err
                )

        return datas

    def _put(self, conn, key, val):
        key = str(key)
        obj = self._new_object(conn, key, val)

        if obj is not None:
            obj.store()

    def _multiput(self, conn, keys, vals):
        keys = list(map(str, keys))
        objs = list(filter(
            lambda obj: obj is not None,
            [
                self._new_object(conn, k, v)
                for k, v in zip(keys, vals)
            ]
        ))

        conn.multiput(objs)

    def _remove(self, conn, key):
        key = str(key)
        obj = self._get_bucket(conn).get(key)

        if not obj.exists:
            raise KeyError('No such key: {0}'.format(key))

        obj.delete()

    def _exists(self, conn, key):
        key = str(key)
        obj = self._get_bucket(conn).get(key)
        return obj.exists

    def _keys(self, conn):
        return self._get_bucket(conn).get_keys()



# -*- coding: utf-8 -*-

from b3j0f.utils.ut import UTCase
from mock import MagicMock, patch
from unittest import main

from riak.datatypes.map import Map as RiakMap
from riak import RiakObject

from link.riak.driver import RiakDriver
from link.crdt.map import Map


class TestDriver(UTCase):
    def setUp(self):
        patcher1 = patch('link.riak.driver.SecurityCreds')
        patcher2 = patch('link.riak.driver.RiakClient')

        self.SecurityCreds = patcher1.start()
        self.RiakClient = patcher2.start()

        self.addCleanup(patcher1.stop)
        self.addCleanup(patcher2.stop)

        self.security = MagicMock()
        self.SecurityCreds.return_value = self.security

        self.conn = MagicMock()
        self.RiakClient.return_value = self.conn

        # middleware configuration
        self.drvconfig = {
            'hosts': [('localhost', 9000)],
            'user': 'guest',
            'pwd': 'guest',
            'default_bucket': 'testbucket',
            'protocol': 'http',
            'pkey': 'private key',
            'cert': 'certificate',
            'cacert': 'certificate authority',
            'crl': 'revocation list',
            'ciphers': 'ssl ciphers',
            'sslver': 5
        }

    def _instanciate(self, protocol='http'):
        self.drvconfig['protocol'] = protocol

        self.drv = RiakDriver(**self.drvconfig)

    def test_connect_nosecurity(self):
        self.drvconfig = {
            'hosts': [('localhost', 9000)]
        }
        self._instanciate()

        self.assertIs(self.conn, self.drv.conn)
        self.SecurityCreds.assert_not_called()
        self.RiakClient.assert_called_with(
            nodes=[
                {
                    'host': 'localhost',
                    'http_port': 9000
                }
            ],
            credentials=None
        )

    def test_connect_http(self):
        self._instanciate(protocol='http')

        self.assertIs(self.conn, self.drv.conn)

        self.SecurityCreds.assert_called_with(
            username='guest',
            password='guest',
            pkey_file='private key',
            cert_file='certificate',
            cacert_file='certificate authority',
            crl_file='revocation list',
            ciphers='ssl ciphers',
            ssl_version=5
        )
        self.RiakClient.assert_called_with(
            nodes=[
                {
                    'host': 'localhost',
                    'http_port': 9000
                }
            ],
            credentials=self.security
        )

    def test_connect_pbc(self):
        self._instanciate(protocol='pbc')

        self.assertIs(self.conn, self.drv.conn)

        self.SecurityCreds.assert_called_with(
            username='guest',
            password='guest',
            pkey_file='private key',
            cert_file='certificate',
            cacert_file='certificate authority',
            crl_file='revocation list',
            ciphers='ssl ciphers',
            ssl_version=5
        )
        self.RiakClient.assert_called_with(
            nodes=[
                {
                    'host': 'localhost',
                    'pb_port': 9000
                }
            ],
            credentials=self.security
        )

    def test_disconnect(self):
        self._instanciate()
        self.drv.connect()

        self.conn.is_alive.return_value = True
        self.assertTrue(self.drv.isconnected())

        self.drv.disconnect()

        self.conn.is_alive.return_value = False
        self.assertFalse(self.drv.isconnected())

        self.conn.close.assert_called_with()

    def test_default_bucket(self):
        bucket = 'expected'

        self.conn.bucket.return_value = bucket
        self._instanciate()
        self.drv.connect()

        result = self.drv._get_bucket(self.drv.conn)

        self.assertEqual(result, bucket)
        self.conn.bucket.assert_called_with('testbucket')

    def test_bucket(self):
        bucket = 'expected'

        self.conn.bucket.return_value = bucket

        self.drvconfig['path'] = ['testbucket2']
        self._instanciate()
        self.drv.connect()

        result = self.drv._get_bucket(self.drv.conn)

        self.assertEqual(result, bucket)
        self.conn.bucket.assert_called_with('testbucket2')

    def test_typed_bucket(self):
        bucket = 'expected'

        bucket_type = MagicMock()
        bucket_type.bucket.return_value = bucket
        self.conn.bucket_type.return_value = bucket_type

        self.drvconfig['path'] = ['testtype', 'testbucket']
        self._instanciate()
        self.drv.connect()

        result = self.drv._get_bucket(self.drv.conn)

        self.assertEqual(result, bucket)
        self.conn.bucket_type.assert_called_with('testtype')
        bucket_type.bucket.assert_called_with('testbucket')

    def test_get(self):
        obj = MagicMock()
        obj.__class__ = RiakObject
        obj.exists = True
        obj.data = 'expected'

        bucket = MagicMock()
        bucket.get.return_value = obj
        self.conn.bucket.return_value = bucket

        self._instanciate()
        self.drv.connect()

        result = self.drv._get(self.drv.conn, 'key')

        self.assertEqual(result, 'expected')
        bucket.get.assert_called_with('key')

    def test_get_crdt(self):
        obj = RiakMap(value={
            ('a', 'counter'): 1
        })

        bucket = MagicMock()
        bucket.get.return_value = obj
        self.conn.bucket.return_value = bucket

        self._instanciate()
        self.drv.connect()

        result = self.drv._get(self.drv.conn, 'key')

        self.assertIsInstance(result, Map)
        self.assertEqual(result.current, {'a_counter': 1})
        bucket.get.assert_called_with('key')

    def test_get_fail(self):
        obj = MagicMock()
        obj.__class__ = RiakObject
        obj.exists = False

        bucket = MagicMock()
        bucket.get.return_value = obj
        self.conn.bucket.return_value = bucket

        self._instanciate()
        self.drv.connect()

        with self.assertRaises(KeyError):
            self.drv._get(self.drv.conn, 'key')

        bucket.get.assert_called_with('key')

        obj = MagicMock()
        bucket.get.return_value = obj

        with self.assertRaises(KeyError):
            self.drv._get(self.drv.conn, 'key')

        bucket.get.assert_called_with('key')

    def test_multiget(self):
        obj0 = MagicMock()
        obj0.__class__ = RiakObject
        obj0.data = 'expected'

        obj1 = RiakMap(value={
            ('a', 'counter'): 1
        })

        bucket = MagicMock()
        bucket.multiget.return_value = [obj0, obj1]
        self.conn.bucket.return_value = bucket

        self._instanciate()
        self.drv.connect()

        result = self.drv._multiget(self.drv.conn, ['foo', 'bar'])

        bucket.multiget.assert_called_with(['foo', 'bar'])
        self.assertEqual(result[0], 'expected')
        self.assertEqual(result[1].current, {'a_counter': 1})

    def test_multiget_fail(self):
        err0 = (None, None, 'foo', Exception('not found'))

        bucket = MagicMock()
        bucket.multiget.return_value = [err0]
        self.conn.bucket.return_value = bucket

        self._instanciate()
        self.drv.connect()

        with self.assertRaises(KeyError):
            self.drv._multiget(self.drv.conn, ['foo', 'bar'])

        bucket.multiget.assert_called_with(['foo', 'bar'])

    def test_put(self):
        bucket = MagicMock()
        obj = MagicMock()
        bucket.new.return_value = obj
        self.conn.bucket.return_value = bucket

        self._instanciate()
        self.drv.connect()

        self.drv._put(self.drv.conn, 'key', {
            'foo': 'bar'
        })

        obj.store.assert_called_with()
        bucket.new.assert_called_with('key', {'foo': 'bar'})

    def test_put_crdt(self):
        bucket = MagicMock()
        self.conn.bucket.return_value = bucket

        self._instanciate()
        self.drv.connect()

        crdt = Map()
        crdt['a_counter'].increment()

        result = self.drv._put(self.drv.conn, 'key', crdt)

        self.assertIsNone(result)
        self.assertTrue(bucket._client.update_datatype.called)
        self.assertEqual(len(bucket._client.update_datatype.call_args), 2)
        self.assertEqual(len(bucket._client.update_datatype.call_args[0]), 1)
        self.assertIsInstance(
            bucket._client.update_datatype.call_args[0][0],
            RiakMap
        )
        self.assertEqual(
            bucket._client.update_datatype.call_args[1],
            {'return_body': True}
        )

    def test_multiput(self):
        bucket = MagicMock()

        self._instanciate()
        self.drv.connect()

        obj = MagicMock()
        crdt = Map()
        crdt['a_counter'].increment()

        bucket.new.return_value = obj
        self.conn.bucket.return_value = bucket

        self.drv._multiput(self.drv.conn, ['foo', 'bar'], [obj, crdt])
        self.conn.multiput.assert_called_with([obj])

    def test_remove(self):
        obj = MagicMock()
        obj.exists = True

        bucket = MagicMock()
        bucket.get.return_value = obj
        self.conn.bucket.return_value = bucket

        self._instanciate()
        self.drv.connect()

        self.drv._remove(self.drv.conn, 'key')

        bucket.get.assert_called_with('key')
        obj.delete.assert_called_with()

    def test_remove_fail(self):
        obj = MagicMock()
        obj.exists = False

        bucket = MagicMock()
        bucket.get.return_value = obj
        self.conn.bucket.return_value = bucket

        self._instanciate()
        self.drv.connect()

        with self.assertRaises(KeyError):
            self.drv._remove(self.drv.conn, 'key')

        bucket.get.assert_called_with('key')
        obj.delete.assert_not_called()

    def test_exists(self):
        obj = MagicMock()

        bucket = MagicMock()
        bucket.get.return_value = obj
        self.conn.bucket.return_value = bucket

        self._instanciate()
        self.drv.connect()

        obj.exists = True
        result = self.drv._exists(self.drv.conn, 'key')
        self.assertTrue(result)
        bucket.get.assert_called_with('key')

        obj.exists = False
        result = self.drv._exists(self.drv.conn, 'key')
        self.assertFalse(result)
        bucket.get.assert_called_with('key')

    def test_keys(self):
        bucket = MagicMock()
        bucket.get_keys.return_value = ['foo', 'bar']
        self.conn.bucket.return_value = bucket

        self._instanciate()
        self.drv.connect()

        result = self.drv._keys(self.drv.conn)
        self.assertEqual(result, ['foo', 'bar'])


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

from link.fulltext.feature import FulltextIndex


class RiakSearch2(FulltextIndex):

    DATA_ID = '_yz_rk'

    def search(self, query):
        result = self.obj.conn.fulltext_search(
            str(self.obj.path[0]),
            query
        )

        return result['docs']



# -*- coding: utf-8 -*-

from b3j0f.utils.ut import UTCase
from unittest import main

from link.crdt.map import TYPES

from link.riak.features.crdt import convert_crdt_from_riak
from link.riak.features.crdt import convert_crdt_to_riak


class TestCRDT(UTCase):
    def test_convert_counter(self):
        initial = TYPES['counter']()
        initial.increment()

        riak = convert_crdt_to_riak(initial)
        self.assertEqual(riak._increment, 1)

        crdt = convert_crdt_from_riak(riak)
        self.assertEqual(crdt._increment, 1)

    def test_convert_flag(self):
        initial = TYPES['flag']()
        initial.enable()

        riak = convert_crdt_to_riak(initial)
        self.assertEqual(riak._op, 'enable')

        crdt = convert_crdt_from_riak(riak)
        self.assertEqual(crdt._mutation, 'enable')

    def test_convert_set(self):
        initial = TYPES['set']()
        initial.add('hello')
        initial.add('world')
        initial.discard('hello')

        riak = convert_crdt_to_riak(initial)
        self.assertEqual(riak._adds, {'hello', 'world'})
        self.assertEqual(riak._removes, {'hello'})

        crdt = convert_crdt_from_riak(riak)
        self.assertEqual(crdt._adds, {'hello', 'world'})
        self.assertEqual(crdt._removes, {'hello'})

    def test_convert_register(self):
        initial = TYPES['register']()
        initial.assign('foo')

        riak = convert_crdt_to_riak(initial)
        self.assertEqual(riak._new_value, 'foo')

        crdt = convert_crdt_from_riak(riak)
        self.assertEqual(crdt._new, 'foo')

    def test_convert_map(self):
        initial = TYPES['map'](value={
            'g_flag': False
        })

        initial['a_counter'].increment()
        initial['b_flag'].enable()
        initial['c_set'].add('hello')
        initial['c_set'].add('world')
        initial['c_set'].discard('hello')
        initial['d_register'].assign('foo')
        initial['e_map']['a_counter'].increment()
        initial['f_counter'].increment()
        del initial['f_counter']

        riak = convert_crdt_to_riak(initial)

        keys = [
            ('a', 'counter'),
            ('b', 'flag'),
            ('c', 'set'),
            ('d', 'register'),
            ('e', 'map'),
            ('f', 'counter')
        ]

        for key in keys:
            self.assertIn(key, riak._updates)

        self.assertIn(('f', 'counter'), riak._removes)
        self.assertIn(('g', 'flag'), riak)
        self.assertEqual(riak[('g', 'flag')].value, False)

        self.assertEqual(riak._updates[('a', 'counter')]._increment, 1)
        self.assertEqual(riak._updates[('b', 'flag')]._op, 'enable')
        self.assertEqual(riak._updates[('c', 'set')]._adds, {'hello', 'world'})
        self.assertEqual(riak._updates[('c', 'set')]._removes, {'hello'})
        self.assertEqual(riak._updates[('d', 'register')]._new_value, 'foo')

        self.assertIn(('a', 'counter'), riak._updates[('e', 'map')]._updates)
        self.assertEqual(
            riak._updates[('e', 'map')]._updates[('a', 'counter')]._increment,
            1
        )

        crdt = convert_crdt_from_riak(riak)

        keys = [
            'a_counter',
            'b_flag',
            'c_set',
            'd_register',
            'e_map',
            'f_counter'
        ]

        for key in keys:
            self.assertIn(key, crdt._updates)

        self.assertIn('f_counter', crdt._removes)
        self.assertIn('g_flag', crdt)
        self.assertEqual(crdt['g_flag'].current, False)

        self.assertEqual(crdt._updates['a_counter']._increment, 1)
        self.assertEqual(crdt._updates['b_flag']._mutation, 'enable')
        self.assertEqual(crdt._updates['c_set']._adds, {'hello', 'world'})
        self.assertEqual(crdt._updates['c_set']._removes, {'hello'})
        self.assertEqual(crdt._updates['d_register']._new, 'foo')

        self.assertIn('a_counter', crdt._updates['e_map']._updates)
        self.assertEqual(
            crdt._updates['e_map']._updates['a_counter']._increment,
            1
        )


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

from b3j0f.utils.ut import UTCase
from mock import MagicMock
from unittest import main

from link.riak.features.fulltext import RiakSearch2


class TestFulltext(UTCase):
    def test_search(self):
        middleware = MagicMock()
        middleware.path = ['test']
        middleware.conn.fulltext_search.return_value = {
            'docs': 'expected'
        }

        rs2 = RiakSearch2(middleware)
        result = rs2.search('query')

        self.assertEqual(result, 'expected')
        middleware.conn.fulltext_search.assert_called_with('test', 'query')


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

from riak.datatypes.datatype import TYPES as RIAK_TYPES
from link.crdt.map import TYPES as LINK_TYPES


class CRDTConverter(object):
    def new_riak_crdt(self, crdt, value, context=None):
        return RIAK_TYPES[crdt._type_name](value=value, context=context)

    def new_link_crdt(self, crdt, value, context=None):
        return LINK_TYPES[crdt.type_name](value=value, context=context)

    def to_counter(self, crdt, context=None):
        obj = self.new_riak_crdt(crdt, crdt._value, context=context)
        obj._increment = crdt._increment
        return obj

    def from_counter(self, crdt, context=None):
        obj = self.new_link_crdt(crdt, crdt.value, context=context)
        obj._increment = crdt._increment
        return obj

    def to_flag(self, crdt, context=None):
        obj = self.new_riak_crdt(crdt, crdt._value, context=context)
        obj._op = str(crdt._mutation) if crdt._mutation is not None else None
        return obj

    def from_flag(self, crdt, context=None):
        obj = self.new_link_crdt(crdt, crdt.value, context=context)
        obj._mutation = crdt._op
        return obj

    def to_set(self, crdt, context=None):
        obj = self.new_riak_crdt(crdt, crdt._value, context=context)
        obj._adds = set(map(str, crdt._adds))
        obj._removes = set(map(str, crdt._removes))
        return obj

    def from_set(self, crdt, context=None):
        obj = self.new_link_crdt(crdt, crdt.value, context=context)
        obj._adds = crdt._adds
        obj._removes = crdt._removes
        return obj

    def to_register(self, crdt, context=None):
        obj = self.new_riak_crdt(crdt, crdt._value, context=context)
        obj._new_value = str(crdt._new) if crdt._new is not None else None
        return obj

    def from_register(self, crdt, context=None):
        obj = self.new_link_crdt(crdt, crdt.value, context=context)
        obj._new = crdt._new_value
        return obj

    def to_map(self, crdt, context=None):
        obj = self.new_riak_crdt(crdt, {}, context=context)

        val = {}

        for key in crdt._value:
            subval = self.to_riak(crdt._value[key], context=obj)
            key, datatype = str(key).rsplit('_', 1)

            val[(key, datatype)] = subval

        obj._value = val

        updates = {}

        for key in crdt._updates:
            subval = self.to_riak(crdt._updates[key], context=obj)
            key, datatype = str(key).rsplit('_', 1)

            updates[(key, datatype)] = subval

        obj._removes = []

        for key in crdt._removes:
            key, datatype = str(key).rsplit('_', 1)
            obj._removes.append((key, datatype))

        obj._updates = updates

        return obj

    def from_map(self, crdt, context=None):
        obj = self.new_link_crdt(crdt, {}, context=context)

        val = {}

        for key, datatype in crdt._value:
            subval = self.from_riak(
                crdt._value[(key, datatype)],
                context=obj
            )
            key = '{0}_{1}'.format(key, datatype)

            val[key] = subval

        obj._value = val

        updates = {}

        for key, datatype in crdt._updates:
            subval = self.from_riak(
                crdt._updates[(key, datatype)],
                context=obj
            )
            key = '{0}_{1}'.format(key, datatype)

            updates[key] = subval

        obj._removes = [
            '{0}_{1}'.format(key, datatype)
            for key, datatype in crdt._removes
        ]

        obj._updates = updates

        return obj

    def to_riak(self, crdt, context=None):
        typemap = {
            'counter': self.to_counter,
            'flag': self.to_flag,
            'set': self.to_set,
            'register': self.to_register,
            'map': self.to_map
        }

        return typemap[crdt._type_name](crdt, context=context)

    def from_riak(self, crdt, context=None):
        typemap = {
            'counter': self.from_counter,
            'flag': self.from_flag,
            'set': self.from_set,
            'register': self.from_register,
            'map': self.from_map
        }

        return typemap[crdt.type_name](crdt, context=context)


def convert_crdt_to_riak(crdt):
    converter = CRDTConverter()
    return converter.to_riak(crdt)


def convert_crdt_from_riak(riak_crdt):
    converter = CRDTConverter()
    return converter.from_riak(riak_crdt)
